/* Mauscon: Nachrichtenformatskonverter fuer das MausNet
   Copyright (C) 1994, 1995, 1996 Uwe Ohse

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


   cancel.c: Dealing with cancel messages 
*/

#include "config.h"
#include "mausconsys.h"

#include <stdio.h>
#include <syslog.h>
#include <stdarg.h>

#include "mauscon.h"
#include "mausconlib.h"

static cancel_t *anker;

cancel_t *
cancel_anker(void)
{
	return anker;
}

int
cancel_init(int keep_locked)
{
	FILE *f;
	int fatal_err=0;
	unsigned int sleeptime=1;
	char buf[512];
	char *p;
	cancel_t *neu;
	uostr_t lfname=UOSTR_INIT;
	uostr_t cfname=UOSTR_INIT;

	uostr_xdup_cstrmulti(&lfname,opts.spoolbase,"/",action,"/cancel.lock",0);
	uostr_x0(&lfname);

	while (lock_create (lfname.data, &fatal_err))
	{
		if (fatal_err)
		{
			log(LOG_ERR,_("cannot create lockfile %s"),lfname.data);
			uostr_freedata(&lfname);
			return 1;
		}
		sleep(sleeptime);
		if (sleeptime<120)
			sleeptime*=2;
	}

	uostr_xdup_cstrmulti(&cfname,opts.spoolbase,"/",action,"/cancel.list",0);
	uostr_x0(&cfname);
	f=fopen(cfname.data,"r");
	if (!f)
	{
		if (!keep_locked)
			lock_delete(lfname.data);
		uostr_freedata(&cfname);
		uostr_freedata(&lfname);
		return 0; /* dann gibt's die Datei halt nicht */
	}
	uostr_freedata(&cfname);
	while(fgets(buf,sizeof(buf),f))
	{
		char *q;
		char *at;
		p=strpbrk(buf,"\r\n");
		if (p)
			*p='\0';
		q=strpbrk(buf,"\t ");
		if (q)
		{
			*q++='\0';
			q=strtrim(q);
		}
		else
			continue;
		p=strtrim(buf);
		at=strchr(p,'@');
		if (!at)
			continue;
		neu=(cancel_t *) XMALLOC(sizeof(cancel_t));
		neu->id=XSTRDUP(p);
		neu->from=XSTRDUP(q);
		neu->done=0;
		neu->next=anker;
		neu->local_part_len=at-p;
		anker=neu;
	}
	fclose(f);
	if (!keep_locked)
		lock_delete(lfname.data);
	uostr_freedata(&lfname);
	return 0;
}

cancel_t * 
cancel_find_in_list(const char *id)
{
	cancel_t *akt;
	char *id_at;
	size_t id_local_l;
	akt=anker;
	id_at=strchr(id,'@');
	if (!id_at)
		return NULL;
	id_local_l=id_at-id;
	while (akt)
	{
		if (akt->local_part_len==id_local_l) {
			if (!strncmp(akt->id,id,id_local_l)
				&& !strcasecmp(akt->id+akt->local_part_len+1,id_at+1))
				return akt;
		}
		akt=akt->next;
	}
	return NULL;
}

void cancel_end(int do_delete, int rewrite)
{
	uostr_t fname=UOSTR_INIT;
	uostr_xdup_cstrmulti(&fname,opts.spoolbase,"/",action,"/cancel.list",0);
	uostr_x0(&fname);

	if (do_delete)
		unlink(fname.data);
	else if (rewrite)
	{
		FILE *f;
		f=fopen(fname.data,"w");
		if (!f)
			log(LOG_ERR,_("cannot open %s: %s"), fname.data,strerror(errno));
		else
		{
			cancel_t *akt;
			akt=anker;
			while (akt)
			{
				fprintf(f,"%s\t%s\n",akt->id,akt->from);
				akt=akt->next;
			}
			fclose(f);
		}
	}
	uostr_xdup_cstrmulti(&fname,opts.spoolbase,"/",action,"/cancel.lock",0);
	uostr_x0(&fname);
	lock_delete(fname.data);
	uostr_freedata(&fname);
}

void cancel_add(const char *id,const char *from)
{
	const char *at;
	cancel_t *neu;
	if (cancel_find_in_list(id))
		return;
	at=strchr(id,'@');
	if (!at)
		return;
	neu=(cancel_t *)XMALLOC(sizeof(cancel_t));
	neu->id=XSTRDUP(id);
	neu->from=XSTRDUP(from);
	neu->next=anker;
	neu->done=0;
	neu->local_part_len=id - at;
	anker=neu;
}

