/* mauscon: Nachrichtenformatskonverter fuer das MausNet
   Copyright (C) 1992 1993 1994 1995 Michael Kraehe, Uwe Ohse

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   getmausval - gibt den Inhalt einer Config-Var auf dem stdout aus 
*/

#include "mausconsys.h"
#include <stdio.h>
#include <syslog.h>
#include "uogetopt.h"
#include "setvars.h"

#include "mausconlib.h"
#include "mauscon.h"

static const char *opt_cnffname;
static const char *opt_value;
static uogetopt_t myopts[]={
	{'a',"action",UOGO_STRING|UOGO_HIDDEN,&action,0,
	"Channel to use (old name)","CHANNEL"},
	{'C',"channel",UOGO_STRING,&action,0,
	"Channel to use","CHANNEL"},
	{'c',"cnffile",UOGO_STRING,&opt_cnffname,0,
	"Configuration file to use","FILE"},
	{'c',"cnf-file",UOGO_STRING|UOGO_HIDDEN,&opt_cnffname,0,
	"Configuration file to use","FILE"},
	{'s',"set",UOGO_CALLBACK,&setvarfn,0,
	"override variables in the configuration file","VAR=VALUE"},
	{'V',"variable",UOGO_STRING,&opt_value,0,
	"variable to look for","NAME"},
	{'V',"var",UOGO_STRING|UOGO_HIDDEN,&opt_value,0,
	"variable to look for","NAME"},
	{'V',"value",UOGO_STRING|UOGO_HIDDEN,&opt_value,0,
	"variable to look for","NAME"},

	{ 0 ,"copyright", UOGO_PRINT|UOGO_HIDDEN,0,0,
	"Copyright (C) 1999 Uwe Ohse\n"
	PACKAGE " comes with ABSOLUTELY NO WARRANTY,\n"
	"to the extent permitted by law.\n"
	"You may redistribute copies of " PACKAGE "\n"
	"under the terms of the GNU General Public License.\n"
	"For more information about these matters, see the file named COPYING.\n"
	"You may redistribute copies of " PACKAGE "\n",0},

	{0,0,0,0,0,0,0} /* --help and --version */
};

int
main (int argc, char **argv)
{
	const char *slash;

	program_name=argv[0];
	slash=strrchr(program_name,'/');
	if (slash)
		program_name=slash+1;

	uogetopt (program_name, PACKAGE, VERSION, &argc, argv, uogetopt_out,
	  "usage: getmausval [options] [channelname]\n"
	  "read a mauscon configuration variable\n",
	  myopts,
	  "\nReport bugs to <uwe-mauscon@bulkmail.ohse.de>");

	if (!action) {
		action=argv[1];
		argv++;
	}

    if (!action)
    {
        log (LOG_ERR, _("channel not given"));
        exit(2);
    }
    else if (argv[1])
    {
        log (LOG_ERR, _("too many arguments"));
        exit (2);
    }

	do_conf(opt_cnffname,action,setvarscount ? setvars :NULL ,setvarscount);

	openlog (program_name, 0, opts.sl_facility);
	setlogmask(LOG_UPTO(opts.sl_priority));

	if (!opt_value)
		dump_conf(stdout);
	else
		printf ("%s\n", get_conf_value (opt_value));

	closelog ();

	exit (0);
}
