#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <netdb.h>
#include <stdlib.h>
#include "mtalib.h"
#include "uostr.h"

const char *mta_errtxt;
const char *mta_hcmssc;
char *mta_hostname;

mta_ret_t
mta_set_hostname(const char *s)
{
	/* 120 chars ought to be enough for a hostname ... */
	char mhn[120];
	if (!s) {
		if (mta_hostname) return MTA_SUCCESS; /* already did it */
		if (-1==gethostname(mhn,sizeof(mhn)-1)) {
			mta_set_err("gethostname failed",0,"(#4.3.0)",errno);return MTA_DEFER;
		}
		s=mhn;
	}
	if (mta_hostname) free(mta_hostname);
	mta_hostname=(char *)strdup(s);
	if (!mta_hostname) {mta_set_err("out of memory",0,"(#4.3.0)",0); return MTA_DEFER;}
	return MTA_SUCCESS;
}

void
mta_set_err(const char *s1,const char *s2, const char *hashcode,int e)
{
	static char buf1[256];
	static char buf2[10];
	size_t l;
	size_t blen;
	if (hashcode) {
		l=strlen(hashcode);
		if (l>=sizeof(buf2)) {
			strcpy(buf1,"internal trouble: mta_set_err failed, hashcode too long");
			strcpy(buf2,"(#4.3.0)");
			goto out;
		}
		strcpy(buf2,hashcode);
	} else
		*buf2=0;
	l=strlen(s1);
	blen=0;
	if (l>=sizeof(buf1)) {
		strcpy(buf1,"internal trouble: mta_set_err failed, s1 too long");
		strcpy(buf2,"(#4.3.0)");
		goto out;
	}
	memcpy(buf1,s1,l);
	blen=l;
	if (s2) {
		l=strlen(s2);
		if (l+blen+1>=sizeof(buf1)) {
			strcpy(buf1,"internal trouble: mta_set_err failed, s2 too long");
			strcpy(buf2,"(#4.3.0)");
			goto out;
		}
		buf1[blen++]=' ';
		memcpy(buf1+blen,s2,l);
		blen+=l;
	}
	if (!e) {
		s2=strerror(e);
		l=strlen(s2);
		if (l+blen+2>=sizeof(buf1)) {
			strcpy(buf1,"internal trouble: mta_set_err failed, s2 too long");
			strcpy(buf2,"(#4.3.0)");
			goto out;
		}
		buf1[blen++]=':';
		buf1[blen++]=' ';
		memcpy(buf1+blen,s2,l);
		blen+=l;
	}
	buf1[blen]=0;
  out:
	mta_errtxt=buf1;
	if (!*buf2) mta_hcmssc=0;
	else mta_hcmssc=buf2;
}
