/* nntp2maus: NNTP->MausTausch-Konverter
   Copyright (C) 1999, 2000 Uwe Ohse

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <unistd.h>

#include <errno.h>
#include <ctype.h>
#include <fcntl.h>
#include <syslog.h>

#if HAVE_DIRENT_H
# include <dirent.h>
# define NAMLEN(dirent) strlen((dirent)->d_name)
#else
# define dirent direct
# define NAMLEN(dirent) (dirent)->d_namlen
# if HAVE_SYS_NDIR_H
#  include <sys/ndir.h>
# endif
# if HAVE_SYS_DIR_H
#  include <sys/dir.h>
# endif
# if HAVE_NDIR_H
#  include <ndir.h>
# endif
#endif

#include <sys/types.h>

#ifdef HAVE_SYS_STAT_H
#  include <sys/stat.h>
#endif

#ifdef HAVE_MMAP
#include <sys/mman.h>
#ifndef MAP_FILE
#define MAP_FILE 0
#endif
#endif

#include "timselsysdep.h"
#include <netinet/in.h>
#include <sys/socket.h>
#include "uosock.h"

#include "mausconlib.h"
#include "mauscon.h"
#include "metamsg.h"
#include "errormail.h"
#include "rfc.h"
#include "maus.h"
#include "groups.h"
#include "domains.h"
#include "rfcaddr.h"
#include "msg_io.h"
#include "uostr.h"
#include "uoio.h"
#include "uogetopt.h"
#include "dynhash.h"
#include "str2num.h"
#include "timing.h"
#include "filelock.h"
#include "crc32.h"
#include "setvars.h"


/* Weil MAUS \r zwingend braucht */
#define LINEEND "\r\n"

static int opt_no_update;
static int opt_timing;
static int opt_statistics;
static int opt_verbose;
static const char *opt_ITG;
static const char *opt_LOG;
static char *opt_outputfilename;

static void open_nntp_server(void);
static long handle_group (const char *groupname, long counter, BOOLEAN *nosuchgroup);
static int handle_group_rc(const char *this_rc, int keep);
static int rewrite_group_rc(const char *rcfname);
static void create_ITG(const char *fname);
static void handle_change(const char *rcfname);

struct group {
	char *name;
	long counter;
	int is_deleted;
	struct group *next;
};
struct group *group_anker;
static void *ghash_handle=0;
static uoio_t sock_i;
static uoio_t sock_o;
static int sockfd=-1;
static mio_t *ochan;


const char *opt_cnffname=SYSCONFDIR "/" CNFNAME;
const char *opt_rcfname=".tauschrc";

static uogetopt_t myopts[] =
{
	{'a',"action",UOGO_STRING|UOGO_HIDDEN,&action,0,
	 "Channel to use (old name)","CHANNEL"},
	{'C',"channel",UOGO_STRING,&action,0,
	 "Channel to use","CHANNEL"},
	{'c',"cnffile",UOGO_STRING,&opt_cnffname,0,
	 "Configuration file to use","FILE"},
	{'c',"cnf-file",UOGO_STRING|UOGO_HIDDEN,&opt_cnffname,0,
	 "Configuration file to use","FILE"},
	{'I',"ITG",UOGO_STRING,&opt_ITG,0,
	 "Create ITG" ,"FILE"},
	{'L',"LOG",UOGO_STRING,&opt_LOG,0,
	 "Create LOG" ,"FILE"},

	{'n', "no-update", UOGO_FLAG, &opt_no_update, 1,
	 "Do change any files",0},
	{'o',"output",UOGO_STRING,&opt_outputfilename,0,
	 "Output file name","FILE"},
	{'r',"rcfile",UOGO_STRING,&opt_rcfname,0,
	 ".tauschrc file\n"
	 "This file contains a list of server names,\n"
	 "newsgroup names and article number high-water-marks\n"
	 "in that groups. Keep off unless you really know what\n"
	 "you are doing (in that case you won't touch that file",
	 "FILE"},
	{'s',"set",UOGO_CALLBACK,&setvarfn,0,
	 "override variables in the configuration file","VAR=VALUE"},
	{'S', "statictics", UOGO_FLAG, &opt_statistics, 1,
	 "Collect statistics.\n"
	 "Such as: bytes + messages transferred.\n",0},
	{'t', "timing", UOGO_FLAG, &opt_timing, 1,
	 "Collect timing information\n"
	 "This is not really useful, except for benchmarks\n"
	 "and profiling",0},
	{'v', "verbose", UOGO_FLAG, &opt_verbose, 1,
	 "Be verbose\n",0},

	{ 0 ,"copyright", UOGO_PRINT|UOGO_HIDDEN,0,0,
	"Copyright (C) 1999,2000 Uwe Ohse\n"
	PACKAGE " comes with ABSOLUTELY NO WARRANTY,\n"
	"to the extent permitted by law.\n"
	"You may redistribute copies of " PACKAGE "\n"
	"under the terms of the GNU General Public License.\n"
	"For more information about these matters, see the file named COPYING.\n"
	"You may redistribute copies of " PACKAGE "\n",0},
	{0, 0, 0, 0, 0, 0, 0}       /* --help and --version */
};

static unsigned long stat_nntp_bytes_read;
static unsigned long stat_nntp_messages_read;
static unsigned long stat_nntp_messages_notread;
static unsigned long stat_nntp_no_cr;
static unsigned long stat_unconverted;
static unsigned long stat_duplicate;

static uoio_t LOG;

int
main (int argc, char **argv)
{
	uostr_t errbuf=UOSTR_INIT;
	int fd;
	unsigned long tim1=0,tim2=0,tim3=0,tim4=0;
	uostr_t oncefname=UOSTR_INIT;
	uostr_t lck=UOSTR_INIT;

	const char *slash;

	start();

	program_name=argv[0];
	slash=strrchr(program_name,'/');
	if (slash)
		program_name=slash+1;

	uogetopt (program_name, PACKAGE, VERSION, &argc, argv, uogetopt_out, 
		"usage: nntp2maus [options] [channelname]\n", myopts,
	     "\nReport bugs to <uwe-mauscon@bulkmail.ohse.de>");
	if (opt_LOG) {
		int logfd=open(opt_LOG,O_WRONLY|O_CREAT|O_APPEND);
		if (logfd==-1) {
			log (LOG_ERR, "cannot open %s: %s",opt_LOG,strerror(errno));
			exit(2);
		}
		uoio_assign_w(&LOG,logfd,write,0);
		uoio_write_cstr(&LOG,":#EXPORT\r\n");
	}

	if (!action) {
		if (!argv[1]) {
			log (LOG_ERR, "channel not given");
			exit(2);
		}
		action=argv[1];
		argv++;
	}
	if (argv[1]) {
		log (LOG_ERR, _("too many arguments"));
		exit (2);
	}

	/* configfile lesen */
    if (do_conf (opt_cnffname, action,setvarscount ? setvars : NULL,setvarscount))
        exit (1);

	openlog (program_name, 0, opts.sl_facility);
	setlogmask(LOG_UPTO(opts.sl_priority));
	open_msglog();

	if (!grp_load_list(FALSE))
		exit (1);				/* Fehlermeldung schon ausgegeben */

	if (!dom_init())
		exit(1);

	/* create output channel */
	if (MIO_OK!=mio_maustausch.open(
		opt_outputfilename ? opt_outputfilename : "-", 
		"w",
		&ochan,
		&errbuf))
	{
		fprintf (stderr, "rfc2maus: %s: %s\n", 
			opt_outputfilename ? opt_outputfilename : "stdin",
			 strerror (errno));
		exit (1);
	}
	uostr_freedata(&errbuf);

	/* create a lock file */
	{
		int tries=0;
		char numbuf[STR_ULONG];
		uostr_xdup_cstr(&lck,opt_rcfname);
		uostr_xadd_mem(&lck,".LCK",5);
		/* read and lock the .tauschrc */
		fd=open(lck.data,O_RDWR|O_CREAT,0644);
		if (fd==-1) {
			log(LOG_ERR,"cannot open %s: %s",lck.data,strerror(errno));
			exit(1);
		}
		while (1) {
			if (tries++==5) {
				log(LOG_ERR,"cannot lock %s: %s",lck.data,strerror(errno));
				exit(1);
			}
			if (fd_write_lock0(fd)) 
				sleep(1);
			else
				break;
		}
		memcpy(numbuf+str_ulong(numbuf,getpid()),"\n",2);
		write(fd,numbuf,strlen(numbuf)); /* just for info */
	}
	open_nntp_server();

	tim1=stop();
	start();

	/* addition to / deletions from the list */
	handle_change(opt_rcfname);

	/* read from that FD, storing the groups and counters */
	handle_group_rc(opt_rcfname, TRUE);

	/* one time jobs */
	uostr_xdup_cstr(&oncefname,opt_rcfname);
	uostr_xadd_cstr(&oncefname,".once");
	uostr_x0(&oncefname);
	/* read it, but don't store into .tauschrc */
	handle_group_rc(oncefname.data, FALSE);
	uostr_freedata(&oncefname);

	/* rewrite groups file, throwing out unwanted groups */
	if (!opt_no_update)
		rewrite_group_rc(opt_rcfname);
	unlink(lck.data);
	uostr_freedata(&lck);

	tim2=stop();
	start();
	if (opt_ITG)
		create_ITG(opt_ITG);
	tim3=stop();
	start();

	ochan->fn->close(ochan);
	closelog ();
	close_msglog();
	uoio_destroy(&sock_i);
	uoio_destroy(&sock_o);
	close(sockfd);

	tim4=stop();

        
	if (opt_statistics) {
		fprintf(stderr,"NNTP: %lu bytes read\n",stat_nntp_bytes_read);
		fprintf(stderr,"NNTP: %lu messages read\n",stat_nntp_messages_read);
		fprintf(stderr,"NNTP: %lu messages not read\n",stat_nntp_messages_notread);
		fprintf(stderr,"NNTP: %lu lines without CR\n",stat_nntp_no_cr);
		fprintf(stderr,"nntp2maus: %lu messages not converted\n",stat_unconverted);
		fprintf(stderr,"nntp2maus: %lu messages duplicate\n",stat_duplicate);
	}
	if (opt_timing) {
		fprintf(stderr,"timing: startup took %lu\n",tim1);
		fprintf(stderr,"timing: work took %lu\n",tim2);
		if (opt_ITG) fprintf(stderr,"timing: ITG took %lu\n",tim3);
		fprintf(stderr,"timing: shutdown took %lu\n",tim4);
	}
	if (opt_LOG) {
		char numbuf[STR_ULONG];
		unsigned long ul;
		uoio_write_cstr(&LOG,":!read ");
		str_ulong(numbuf,stat_nntp_bytes_read);
		uoio_write_cstr(&LOG,numbuf);
		uoio_write_cstr(&LOG," bytes == ");
		str_ulong(numbuf,stat_nntp_messages_read);
		uoio_write_cstr(&LOG,numbuf);
		uoio_write_cstr(&LOG," messages. ");
		str_ulong(numbuf,stat_nntp_messages_notread);
		uoio_write_cstr(&LOG,numbuf);
		uoio_write_cstr(&LOG," holes.\r\n:!");
		str_ulong(numbuf,stat_unconverted);
		uoio_write_cstr(&LOG,numbuf);
		uoio_write_cstr(&LOG," unconverted. ");
		str_ulong(numbuf,stat_duplicate);
		uoio_write_cstr(&LOG,numbuf);
		uoio_write_cstr(&LOG," duplicate.\r\n");
		if (stat_nntp_no_cr) {
			str_ulong(numbuf,stat_nntp_no_cr);
			uoio_write_cstr(&LOG,":!");
			uoio_write_cstr(&LOG,numbuf);
			uoio_write_cstr(&LOG," lines without CR.\r\n");
		}
		uoio_write_cstr(&LOG,":!timing: ");
		str_ulong(numbuf,tim1);
		uoio_write_cstr(&LOG,numbuf);
		uoio_write_cstr(&LOG,", ");
		str_ulong(numbuf,tim2);
		uoio_write_cstr(&LOG,numbuf);
		uoio_write_cstr(&LOG,", ");
		str_ulong(numbuf,tim3);
		uoio_write_cstr(&LOG,numbuf);
		uoio_write_cstr(&LOG,", ");
		str_ulong(numbuf,tim4);
		uoio_write_cstr(&LOG,numbuf);

		if (tim2<1000) tim2=1000;
		ul=(stat_nntp_messages_read-stat_unconverted-stat_duplicate)*1000/(tim2/1000);
		uoio_write_cstr(&LOG," -> ");
		str_ulong(numbuf,ul);
		uoio_write_cstr(&LOG,numbuf);
		uoio_write_cstr(&LOG,"/sec.\r\n");
	}
	if (opt_LOG && -1==uoio_flush(&LOG)) {
		log(LOG_ERR,"cannot flush %s: %s\n",opt_LOG,strerror(errno));
		exit(1);
	}
	exit (0);
}

static void
del_group(const char *name)
{
	struct group *g;
	DYN_ENTRY item,*found;
	item.crc=crc32ansi(name,strlen(name));
	found=hhsearch(ghash_handle,&item,DYN_FIND);
	if (found) {
		g=found->u.ptr;
		g->is_deleted=1;
		return;
	}
	g=XMALLOC(sizeof(struct group));
	g->name=XSTRDUP(name);
	g->is_deleted=1;
	g->counter=0;
	g->next=group_anker;
	group_anker=g;
	item.u.ptr=g;
	if (!hhsearch(ghash_handle,&item,DYN_ENTER)) {
		XNULLCHECK(NULL);
	}
}

static void
add_group(const char *name, long num, int force)
{
	struct group *g;
	DYN_ENTRY item,*found;
	item.crc=crc32ansi(name,strlen(name));
	found=hhsearch(ghash_handle,&item,DYN_FIND);
	if (found) {
		g=found->u.ptr;
		if (g->counter < num || force)
			g->counter=num;
		return;
	}
	g=XMALLOC(sizeof(struct group));
	g->name=XSTRDUP(name);
	g->is_deleted=0;
	g->counter=num;
	g->next=group_anker;
	group_anker=g;
	item.u.ptr=g;
	if (!hhsearch(ghash_handle,&item,DYN_ENTER)) {
		XNULLCHECK(NULL);
	}
}

static void
handle_change(const char *fname)
{
	int fd;
	uostr_t chfname=UOSTR_INIT;
	uoio_t i;
	
	if (!ghash_handle) {
		ghash_handle=hhcreate(513);
		XNULLCHECK(ghash_handle);
	}

	uostr_xdup_cstr(&chfname,fname);
	uostr_xadd_mem(&chfname,".change",8);
	fd=open(chfname.data,O_RDONLY);
	if (fd==-1 && errno==ENOENT) {
		uostr_freedata(&chfname);
		return;
	}
	if (fd==-1) {
		log(LOG_ERR,"cannot open %s: %s",chfname.data,strerror(errno));
		exit(1);
	}
	if (fd!=-1) {
		uoio_assign_r(&i,fd,read,0);
		while (1) {
			ssize_t got;
			char *p;
			got=uoio_getdelim_zc(&i,&p,'\n');
			if (got==-1) {
				log(LOG_ERR,"cannot read %s: %s",chfname.data,strerror(errno));
				exit(1);
			}
			if (got==0) break; /* EOF */
			if (p[got-1]!='\n') {
				log(LOG_ERR,"ignoring unterminated last line in %s",chfname.data);
				break;
			}
			p[got-1]=0;
			if (*p=='+') {
				/* add to group list */
				/* + SPC? NAME SPC NUM */
				/* + SPC? NAME */
				char *ws;
				char *gn;
				long l;
				BOOLEAN nosuchgroup;
				gn=p+1;
				while(*gn && isspace(*gn)) gn++;
				if (!*gn) {
					log(LOG_ERR,"ignoring unknown line in %s: %s",chfname.data,p);
					continue;
				}
				ws=gn;
				while(*ws && !isspace(*ws)) ws++;
				if (*ws) {
					*ws=0;
					ws++;
				}
				while(*ws && isspace(*ws)) ws++;
				if (!*ws) l=-100;
				else {
					size_t x;
					x=str2long(ws,&l,10);
					if (x!=strlen(ws)) {
						log(LOG_ERR,"ignoring unknown count for %s: %s",gn,ws);
						continue;
					}
				}
				l=handle_group (gn,l,&nosuchgroup);
				if (!nosuchgroup)
					add_group(gn,l,1);
			} else if (*p=='-') {
				/* delete group from list */
				/* - SPC? GROUP */
				p++;
				while(*p && isspace(*p)) p++;
				del_group(p);
			} else {
				log(LOG_ERR,"ignoring unknown line in %s: %s",chfname.data,p);
			}
		}
		uoio_destroy(&i);
		close(fd);
		if (!opt_no_update)
			unlink(chfname.data);
	}
	uostr_freedata(&chfname);
}

static int 
rewrite_group_rc(const char *fname)
{
	int fd;
	uostr_t newrc=UOSTR_INIT;
	uostr_t alt=UOSTR_INIT;
	uostr_t neu=UOSTR_INIT;
	
	if (!ghash_handle)
		return 0; /* 0 groups, 0 rewrite */

	/* no write every group in the list. Note that groups which are
	 * not wanted anymore have NULL as name.
	 */
	uostr_xdup_cstr(&newrc,fname);
	uostr_xadd_cstr(&newrc,".new");
	uostr_x0(&newrc);
	fd=open(newrc.data,O_WRONLY|O_CREAT|O_TRUNC,0644);
	if (fd==-1) {
		log(LOG_ERR,"cannot create %s: %s",newrc.data,strerror(errno));
		exit (1);
	} else {
		uoio_t o;
		struct group *g;
		char numbuf[STR_ULONG+1];
		*numbuf=' ';
		uoio_assign_w(&o,fd,write,0);
		for (g=group_anker;g;g=g->next) {
			if (!g->is_deleted) {
				uoio_write_cstr(&o,g->name);
				str_ulong(numbuf+1,g->counter);
				uoio_write_cstr(&o,numbuf);
				uoio_write_mem(&o,"\n",1);
			}
		}
		if (-1==uoio_flush(&o)) {
			log(LOG_ERR,"cannot write to %s: %s",newrc.data,strerror(errno));
			unlink(newrc.data);
			exit(1);
		}
		uoio_destroy(&o);
		close(fd);
	}
	/* create backups */
	{
		int j;
		for (j=2;j>=0;j--) {
			char numbuf[STR_ULONG];
			uostr_xdup_cstr(&alt,opt_rcfname);
			if (j>0) {
				uostr_xadd_mem(&alt,".",1);
				str_ulong(numbuf,j);
				uostr_xadd_cstr(&alt,numbuf);
			}
			uostr_x0(&alt);
			uostr_xdup_cstr(&neu,opt_rcfname);
			uostr_xadd_mem(&neu,".",1);
			str_ulong(numbuf,j+1);
			uostr_xadd_cstr(&neu,numbuf);
			uostr_x0(&neu);
			unlink(neu.data);
			if (-1==rename(alt.data,neu.data) && errno!=ENOENT) {
				log(LOG_ERR,"cannot rename %s to %s: %s",alt.data,neu.data,strerror(errno));
				exit(1);
			}
		}
	}
	/* newrc: freshly written
	 * alt: the .rc file itself, already moved out of the way */
	if (-1==rename(newrc.data,alt.data) && errno!=ENOENT) {
		log(LOG_ERR,"cannot rename %s to %s: %s",newrc.data,alt.data,strerror(errno));
		exit(1);
	}
	uostr_freedata(&alt);
	uostr_freedata(&neu);
	uostr_freedata(&newrc);
	return 0;
}

static int
handle_group_rc(const char *this_rc, int keep)
{
	uoio_t io_i;
	int rcfd;
	if (!ghash_handle) {
		ghash_handle=hhcreate(513);
		XNULLCHECK(ghash_handle);
	}

	rcfd=open(this_rc,O_RDONLY,0644);
	if (rcfd==-1) {
		if (errno==ENOENT)
			return -1;
		log(LOG_ERR,"cannot open %s: %s",this_rc, strerror(errno));
		exit(1);
	}

	uoio_assign_r(&io_i,rcfd,read,0);
	while (1) {
		ssize_t got;
		char *ptr;
		unsigned long ul;
		char *nptr;
		BOOLEAN nosuchgroup;
		got=uoio_getdelim_zc(&io_i,&ptr,'\n');
		if (got==0) /* eof */
			break;
		if (got==-1) { /* error */
			log(LOG_ERR,"cannot read %s: %s",this_rc,strerror(errno));
			exit(1);
		}
		if (ptr[got-1]!='\n') {
			log(LOG_ERR,"unterminated last line in %s ignored",this_rc);
			break;
		}
		ptr[got-1]=0;
		/* Format:
		 *  # comment
		 * group SPACE number
		 */
		while (*ptr && (*ptr==' ' || *ptr=='\t')) ptr++;
		if (*ptr=='#') continue;
		if (!*ptr) {
			log(LOG_ERR,"illegal line in %s ignored",this_rc);
			continue;
		}
		nptr=ptr;
		/* skip to next space */
		while (*nptr && *nptr!=' ' && *nptr!='\t') nptr++;
		if (!*nptr) {
			log(LOG_ERR,"illegal line in %s ignored",this_rc);
			continue;
		}
		*nptr++=0;
		/* skip blanks */
		while (*nptr && (*nptr==' ' || *nptr=='\t')) nptr++;
		if (*nptr=='#') continue;
		if (!*nptr) {
			log(LOG_ERR,"illegal line in %s ignored",this_rc);
			continue;
		}
		if (str2ulong(nptr,&ul,10) < 1) {
			log(LOG_ERR,"illegal number %s in %s ignored",nptr, this_rc);
			continue;
		}
		ul=handle_group(ptr,ul,&nosuchgroup);
		if (keep && !nosuchgroup) {
			add_group(ptr, ul, 0);
		}
	}
	uoio_destroy(&io_i);
	close(rcfd);
	return 0;
}

static ssize_t
read_nntp_answer(char **optr)
{
	ssize_t got;
	char *ptr;
	got=uoio_getdelim_zc(&sock_i,&ptr,'\n');
	if (got==-1) {
		log(LOG_ERR,"can't read from NNTP server: %s",strerror(errno));
		exit(1);
	}
	stat_nntp_bytes_read+=got;
	if (got==0) {
		log(LOG_ERR,"NNTP server closed connection");
		exit(1);
	}
	if (ptr[got-1]!='\n') {
		log(LOG_ERR,"NNTP server didn't send LF");
		exit(1);
	}
	if (got==1 || ptr[got-2]!='\r') {
		/* i've such an article in my database,
		   X-Header: bla\n
		      continue\r\n
		 */
		stat_nntp_no_cr++;
		ptr[got-1]=0;
		got--;
	} else {
		ptr[got-2]=0;
		got-=2;
	}
	*optr=ptr;
	return got;
}

static void
open_nntp_server(void)
{
	struct sockaddr_in s_in;
	const char *server=getenv("NNTPSERVER");
	ssize_t got;
	char *ptr;
	s_in.sin_family=AF_INET;
	s_in.sin_port=htons(119);
	s_in.sin_addr.s_addr=INADDR_LOOPBACK;
	if (!server) server="localhost";
	if (-1==parse_ip_port(server,&s_in.sin_addr,&s_in.sin_port,"tcp")) {
		log(LOG_ERR,"can't parse NNTP server specification %s",server);
		exit(1);
	}
	sockfd=uosock_connect(&s_in,60);
	if (sockfd==-1) {
		log(LOG_ERR,"can't connect to %s: %s",server,strerror(errno));
		exit(1);
	}
	uoio_assign_r(&sock_i,sockfd,read,0);
	uoio_assign_w(&sock_o,sockfd,write,0);
	sock_i.timeout=60;
	sock_o.timeout=60;

	got=read_nntp_answer(&ptr);
	if (*ptr!='2') {
		log(LOG_ERR,"%s doesn't like me: %s",server,ptr);
		exit(1);
	}
	uoio_write_cstr(&sock_o,"MODE READER\r\n");
	uoio_flush(&sock_o);
	/* read "mode reader" answer, and ignore it. Error messages are quite
	 * normal, MODE READER is INN specific.
	 */
	(void) read_nntp_answer(&ptr);
}

#define PRELOAD 4
#if 0
static long
handle_group (const char *groupname, long counter, BOOLEAN *nosuchgroup)
{
	long high,low;
	const char *optr;
	char *ptr;
	long i;
	int preloaded;
	static uostr_t mbuf=UOSTR_INIT;
	metamsg_t *meta;
	static void *hash_handle=0;
	ssize_t got;
	unsigned int verbcount=0; 
	long exportcount=0; 
	int gname_printed=0;
	char numbuf[STR_ULONG];
	if (nosuchgroup)
		*nosuchgroup=FALSE;

	if (!hash_handle) {
		hash_handle=hhcreate(513);
		XNULLCHECK(hash_handle);
	}
	uoio_write_cstr(&sock_o,"GROUP ");
	uoio_write_cstr(&sock_o,groupname);
	uoio_write_cstr(&sock_o,"\r\n");
	uoio_flush(&sock_o);
	got=read_nntp_answer(&ptr);
	if (*ptr!='2') {
		log(LOG_ERR,"can't enter group %s: %s",groupname,ptr);
		if (ptr[0]=='4' && ptr[1]=='1' && ptr[2]=='1') { /* no such group */
			if (nosuchgroup)
				*nosuchgroup=TRUE;
		}
		return counter;
	}

	optr=ptr;
	/* 211 69 210 288 maus.net.gate */
	while (*ptr && !isspace(*ptr)) ptr++; /* skip 211 */
	while (*ptr && isspace(*ptr)) ptr++; /* skip " " */
	while (*ptr && !isspace(*ptr)) ptr++; /* skip 69 */
	while (*ptr && isspace(*ptr)) ptr++; /* skip " " */
	if (str2ulong(ptr,&low,10) < 1) {
		log(LOG_ERR,"can't parse low ptr in group %s: %s",groupname,optr);
		return counter;
	}
	while (*ptr && !isspace(*ptr)) ptr++; /* skip 210 */
	while (*ptr && isspace(*ptr)) ptr++; /* skip " " */
	if (str2ulong(ptr,&high,10) < 1) {
		log(LOG_ERR,"can't parse high ptr in group %s: %s",groupname,optr);
		return counter;
	}


	if (counter < 0) {
		i=high+counter+1;
		if (i<low) i=low;
	} else
		i=counter+1;
	if (i<low) {
		str_ulong(numbuf,low-i);
		if (opt_verbose)
			printf("%s articles lost in %s\n",numbuf,groupname);
		if (opt_LOG) {
			uoio_write_cstr(&LOG,":!");
			uoio_write_cstr(&LOG,numbuf);
			uoio_write_cstr(&LOG," Artikel in ");
			uoio_write_cstr(&LOG,groupname);
			uoio_write_cstr(&LOG," verloren\r\n");
		}
		i=low;
	}
	if (i>high || high==0)
		return high;

	meta=meta_alloc();
	XNULLCHECK(meta);

	for (preloaded=0;preloaded<PRELOAD;preloaded++) {
		if (i+preloaded<=high) {
			uoio_write_cstr(&sock_o,"ARTICLE ");
			str_ulong(numbuf,i+preloaded);
			uoio_write_cstr(&sock_o,numbuf);
			uoio_write_cstr(&sock_o,"\r\n");
		}
	}
	uoio_flush(&sock_o);

	while (i<=high) {
		DYN_ENTRY item;
		DYN_ENTRY *found;
		BOOLEAN in_header=TRUE;
		/* send ARTICLE command */
		if (i+preloaded<=high) {
			uoio_write_cstr(&sock_o,"ARTICLE ");
			str_ulong(numbuf,i+preloaded);
			uoio_write_cstr(&sock_o,numbuf);
			uoio_write_cstr(&sock_o,"\r\n");
			uoio_flush(&sock_o);
		}
		i++;

#define VERBNUM 100
		if (opt_verbose) {
			verbcount++;
			if (!gname_printed) {
				fprintf(stderr,"%s: ",groupname);
				gname_printed=1;
			}
			if (verbcount==VERBNUM/4) {
				fprintf(stderr,"\010/");
			} else if (verbcount==VERBNUM/2) {
				fprintf(stderr,"\010|");
			} else if (verbcount==VERBNUM/4*3) {
				fprintf(stderr,"\010\\");
			} else if (verbcount==VERBNUM) {
				fprintf(stderr,"\010-");
				verbcount=0;
			}
		}

		got=read_nntp_answer(&ptr);
		if (*ptr!='2') {
			if (ptr[0]!='4' || ptr[1]!='2' || ptr[2]!='3')
				log(LOG_ERR,"can't read article %lu in group %s: %s",i-1,groupname,ptr);
			stat_nntp_messages_notread++;
			continue;
		}
		stat_nntp_messages_read++;
		mbuf.len=0;
		while (1) {
			got=read_nntp_answer(&ptr);
			if (got==1 && *ptr=='.')
				break;
			if (got==0) in_header=0;
			if (in_header && (*ptr==' ' || *ptr=='\t')) {
				uostr_cut(&mbuf,-1);
			}
			if (mbuf.len+got>mbuf.size)
				uostr_xadd_mem(&mbuf,ptr,got);
			else {
				memcpy(mbuf.data+mbuf.len,ptr,got);
				mbuf.len+=got;
			}

			if (mbuf.len==mbuf.size)
				uostr_xadd_mem(&mbuf,"\n",1);
			else
				mbuf.data[mbuf.len++]='\n';
		}
		uostr_x0(&mbuf);
	    if (rfc_to_meta(&mbuf,meta,0)) { /* 0: is_news */
			stat_unconverted++;
	        /* Fehler beim Umwandeln */
			continue;
		}
		if (!meta->message_id) {
			log(LOG_ERR,"no message_id header");
			stat_unconverted++;
			continue;
		}
		if (!meta->head_from || !meta->head_from->text) {
			log(LOG_ERR,"no From header: %lu",i-1);
			stat_unconverted++;
			continue;
		}
		if (meta->tm_date==NULL) {
			log(LOG_ERR,"no Date header");
			stat_unconverted++;
			continue;
		}
		if (!meta->newsgroups) {
			log(LOG_ERR,"no newsgroups header");
			stat_unconverted++;
			continue;
		}
	    item.crc=crc32ansi(meta->message_id,strlen(meta->message_id));
		found=hhsearch(hash_handle,&item,DYN_FIND);
		if (found) {
			stat_duplicate++;
			continue; 
		}
		item.u.ptr=NULL;
		if (!hhsearch(hash_handle,&item,DYN_ENTER)) {
			XNULLCHECK(NULL);
		}
		ochan->fn->export(ochan,meta);
		if (opt_verbose)
			exportcount++;
	}
	if (opt_verbose && exportcount)
		fprintf(stderr,"\010 %lu\n",exportcount);
	meta_free(meta);
	return high;
}
#else
static long
handle_group (const char *groupname, long counter, BOOLEAN *nosuchgroup)
{
	long high,low;
	const char *optr;
	char *ptr;
	long i;
	static uostr_t mbuf=UOSTR_INIT;
	metamsg_t *meta;
	static void *hash_handle=0;
	ssize_t got;
	unsigned int verbcount=0; 
	long exportcount=0; 
	int gname_printed=0;
	char numbuf[STR_ULONG];
	long *block;
	int blockcounter=0;
	int j;
	if (nosuchgroup)
		*nosuchgroup=FALSE;

	if (!hash_handle) {
		hash_handle=hhcreate(513);
		XNULLCHECK(hash_handle);
	}
	uoio_write_cstr(&sock_o,"GROUP ");
	uoio_write_cstr(&sock_o,groupname);
	uoio_write_cstr(&sock_o,"\r\n");
	uoio_flush(&sock_o);
	got=read_nntp_answer(&ptr);
	if (*ptr!='2') {
		log(LOG_ERR,"can't enter group %s: %s",groupname,ptr);
		if (ptr[0]=='4' && ptr[1]=='1' && ptr[2]=='1') { /* no such group */
			if (nosuchgroup)
				*nosuchgroup=TRUE;
		}
		return counter;
	}

	optr=ptr;
	/* 211 69 210 288 maus.net.gate */
	while (*ptr && !isspace(*ptr)) ptr++; /* skip 211 */
	while (*ptr && isspace(*ptr)) ptr++; /* skip " " */
	while (*ptr && !isspace(*ptr)) ptr++; /* skip 69 */
	while (*ptr && isspace(*ptr)) ptr++; /* skip " " */
	if (str2ulong(ptr,&low,10) < 1) {
		log(LOG_ERR,"can't parse low ptr in group %s: %s",groupname,optr);
		return counter;
	}
	while (*ptr && !isspace(*ptr)) ptr++; /* skip 210 */
	while (*ptr && isspace(*ptr)) ptr++; /* skip " " */
	if (str2ulong(ptr,&high,10) < 1) {
		log(LOG_ERR,"can't parse high ptr in group %s: %s",groupname,optr);
		return counter;
	}


	if (counter < 0) {
		i=high+counter+1;
		if (i<low) i=low;
	} else
		i=counter+1;
	if (i<low) {
		str_ulong(numbuf,low-i);
		if (opt_verbose)
			printf("%s articles lost in %s\n",numbuf,groupname);
		if (opt_LOG) {
			uoio_write_cstr(&LOG,":!");
			uoio_write_cstr(&LOG,numbuf);
			uoio_write_cstr(&LOG," Artikel in ");
			uoio_write_cstr(&LOG,groupname);
			uoio_write_cstr(&LOG," verloren\r\n");
		}
		i=low;
	}
	if (i>high || high==0)
		return high;

	meta=meta_alloc();
	XNULLCHECK(meta);

	block=XMALLOC( (high-i+1)*sizeof(long));
	blockcounter=0;
#ifdef XPAT
	uoio_write_cstr(&sock_o,"XPAT message-id ");
	str_ulong(numbuf,i);
	uoio_write_cstr(&sock_o,numbuf);
	str_ulong(numbuf,high);
	uoio_write_cstr(&sock_o,"-");
	uoio_write_cstr(&sock_o,numbuf);
	uoio_write_cstr(&sock_o," *\r\n");
	uoio_flush(&sock_o);
	got=read_nntp_answer(&ptr);
	if (*ptr!='2') {
		log(LOG_ERR,"can't XPAT %s: %s",groupname,ptr);
		return counter;
	}
	mbuf.len=0;
	while (1) {
		long l,x;
		got=read_nntp_answer(&ptr);
		if (got==1 && *ptr=='.')
			break;
		x=str2long(ptr,&l,10);
		if (!isspace(ptr[x])) {
			log(LOG_ERR,"can't parse XPAT output: %s",ptr);
			continue;
		}
		if (l>=i)
			block[blockcounter++]=l;
	}
#else
	uoio_write_cstr(&sock_o,"LISTGROUP ");
	uoio_write_cstr(&sock_o,groupname);
	uoio_write_cstr(&sock_o,"\r\n");
	uoio_flush(&sock_o);
	got=read_nntp_answer(&ptr);
	if (*ptr!='2') {
		log(LOG_ERR,"can't LISTGROUP %s: %s",groupname,ptr);
		return counter;
	}
	mbuf.len=0;
	while (1) {
		long l,x;
		got=read_nntp_answer(&ptr);
		if (got==1 && *ptr=='.')
			break;
		x=str2long(ptr,&l,10);
		if (ptr[x]) {
			log(LOG_ERR,"can't parse LISTGROUP output: %s",ptr);
			continue;
		}
		if (l>=i)
			block[blockcounter++]=l;
	}
#endif
	i=0;
	j=0;

	for (i=0;i<PRELOAD;i++) {
		if (i<blockcounter) {
			uoio_write_cstr(&sock_o,"ARTICLE ");
			str_ulong(numbuf,block[i]);
			uoio_write_cstr(&sock_o,numbuf);
			uoio_write_cstr(&sock_o,"\r\n");
		}
	}
	uoio_flush(&sock_o);

	while (j<blockcounter) {
		DYN_ENTRY item;
		DYN_ENTRY *found;
		BOOLEAN in_header=TRUE;
		/* send ARTICLE command */
		if (i<blockcounter) {
			uoio_write_cstr(&sock_o,"ARTICLE ");
			str_ulong(numbuf,block[i]);
			uoio_write_cstr(&sock_o,numbuf);
			uoio_write_cstr(&sock_o,"\r\n");
			uoio_flush(&sock_o);
		}
		i++;
		j++;

#define VERBNUM 100
		if (opt_verbose) {
			verbcount++;
			if (!gname_printed) {
				fprintf(stderr,"%s: ",groupname);
				gname_printed=1;
			}
			if (verbcount==VERBNUM/4) {
				fprintf(stderr,"\010/");
			} else if (verbcount==VERBNUM/2) {
				fprintf(stderr,"\010|");
			} else if (verbcount==VERBNUM/4*3) {
				fprintf(stderr,"\010\\");
			} else if (verbcount==VERBNUM) {
				fprintf(stderr,"\010-");
				verbcount=0;
			}
		}

		got=read_nntp_answer(&ptr);
		if (*ptr!='2') {
			if (ptr[0]!='4' || ptr[1]!='2' || ptr[2]!='3')
				log(LOG_ERR,"can't read article %lu in group %s: %s",i-1,groupname,ptr);
			stat_nntp_messages_notread++;
			continue;
		}
		stat_nntp_messages_read++;
		mbuf.len=0;
		while (1) {
			got=read_nntp_answer(&ptr);
			if (got==1 && *ptr=='.')
				break;
			if (got==0) in_header=0;
			if (in_header && (*ptr==' ' || *ptr=='\t')) {
				uostr_cut(&mbuf,-1);
			}
			uostr_xadd_mem(&mbuf,ptr,got);
			uostr_xadd_mem(&mbuf,"\n",1);
		}
		uostr_x0(&mbuf);
	    if (rfc_to_meta(&mbuf,meta,0)) { /* 0: is_news */
			stat_unconverted++;
	        /* Fehler beim Umwandeln */
			continue;
		}
		if (!meta->message_id) {
			log(LOG_ERR,"no message_id header");
			stat_unconverted++;
			continue;
		}
		if (!meta->head_from || !meta->head_from->text) {
			log(LOG_ERR,"no From header: %lu",i-1);
			stat_unconverted++;
			continue;
		}
		if (meta->date==0 || meta->date==(uo_sec70_t)-1) {
			log(LOG_ERR,"no Date header");
			stat_unconverted++;
			continue;
		}
		if (!meta->newsgroups) {
			log(LOG_ERR,"no newsgroups header");
			stat_unconverted++;
			continue;
		}
	    item.crc=crc32ansi(meta->message_id,strlen(meta->message_id));
		found=hhsearch(hash_handle,&item,DYN_FIND);
		if (found) {
			stat_duplicate++;
			continue; 
		}
		item.u.ptr=NULL;
		if (!hhsearch(hash_handle,&item,DYN_ENTER)) {
			XNULLCHECK(NULL);
		}
		ochan->fn->export(ochan,meta);
		if (opt_verbose)
			exportcount++;
	}
	if (opt_verbose && exportcount)
		fprintf(stderr,"\010 %lu\n",exportcount);
	meta_free(meta);
	free(block);
	return high;
}
#endif

struct itg {
	char *name;
	char *ui;
	char *flags;
	struct itg *next;
};
struct itg *itg_anker;

static unsigned long itg_crc;
static ssize_t crc_write(int fd, const void *buf, size_t count)
{	
	crc32ansi_update(&itg_crc,buf,count);
	return write(fd,buf,count);
	
}


static void
create_ITG(const char *fname)
{
	void *hash_handle=0;
	struct itg *itg;
	char *ptr;
	ssize_t got;
	uoio_t o;
	int fd;
	uostr_t crc_fname=UOSTR_INIT;
	char numbuf[STR_ULONG];

	hash_handle=hhcreate(513);
	XNULLCHECK(hash_handle);

	uoio_write_cstr(&sock_o,"LIST ACTIVE\r\n");
	uoio_flush(&sock_o);
	got=read_nntp_answer(&ptr);
	if (*ptr!='2') {
		log(LOG_ERR,"can't read active file: %s",ptr);
		exit(1);
	}
	while (1) {
		char *p;
		DYN_ENTRY item,*found;

		got=read_nntp_answer(&ptr);
		if (got==1 && *ptr=='.')
			break;
		p=ptr;
		while (*p && !isspace(*p)) p++;
		if (!*p || p==ptr) {
			log(LOG_ERR,"garbled line in active file: %s",ptr);
			exit(1);
		}
		*p=0;
		/* NAME LOW HIGH FLAGS */
		p++;
		while (*p && isspace(*p)) p++;
		while (*p && !isspace(*p)) p++;
		while (*p && isspace(*p)) p++;
		while (*p && !isspace(*p)) p++;
		while (*p && isspace(*p)) p++;

	    item.crc=crc32ansi(ptr,strlen(ptr));
		found=hhsearch(hash_handle,&item,DYN_FIND);
		if (found)
			continue;
		itg=XMALLOC(sizeof(struct itg));
		memset(itg,0,sizeof(struct itg));
		itg->name=XSTRDUP(ptr);
		itg->flags=XSTRDUP(p);
		itg->next=itg_anker;
		itg_anker=itg;
		item.u.ptr=itg;
		if (!hhsearch(hash_handle,&item,DYN_ENTER)) {
			XNULLCHECK(NULL);
		}
	}
	uoio_write_cstr(&sock_o,"LIST NEWSGROUPS\r\n");
	uoio_flush(&sock_o);
	got=read_nntp_answer(&ptr);
	if (*ptr!='2') {
		log(LOG_ERR,"can't read newsgroups file: %s",ptr);
		exit(1);
	}
	while (1) {
		char *p;
		DYN_ENTRY item,*found;

		got=read_nntp_answer(&ptr);
		if (got==1 && *ptr=='.')
			break;
		p=ptr;
		while (*p && !isspace(*p)) p++;
		if (!*p || p==ptr) {
			log(LOG_ERR,"garbled line in newsgroups file: %s",ptr);
			exit(1);
		}
		*p=0;
		p++;
		while (*p && isspace(*p)) p++;
		if (!*p) continue;

	    item.crc=crc32ansi(ptr,strlen(ptr));
		found=hhsearch(hash_handle,&item,DYN_FIND);
		if (!found)
			continue;
		itg=found->u.ptr;
		itg->ui=XSTRDUP(p);
	}
	fd=open(fname,O_WRONLY|O_CREAT|O_TRUNC,0644);
	if (fd==-1) {
		log(LOG_ERR,"can't open %s: %s",fname,strerror(errno));
		exit(1);
	}
	crc32ansi_init(&itg_crc);
	uoio_assign_w(&o,fd,crc_write,0);
	for(itg=itg_anker;itg;itg=itg->next) {
		grp_data_t *g;

		/* + GMaus.Sysop */
		/* - ASYSOPS */
		/* + UInterne Diskussion von MAUS-SysOps */
		/* / "COlaf Boos" -> "CSysop" */
		/* FL+S+V+G+P-$- */
		/* - MG */
		/* / H1:MausNet */
		/* / P1:Deutsch */
		g=grp_suche_mausname(itg->name,0);
		uoio_write_mem(&o,"G",1);
		uoio_write_cstr(&o,(g && g->mausname) ? g->mausname : itg->name);
		uoio_write_mem(&o,"\r\n",2);
		uoio_write_mem(&o,"U",1);
		uoio_write_cstr(&o,itg->ui ? itg->ui : "Keine Beschreibung verfgbar");
		uoio_write_mem(&o,"\r\n",2);
		uoio_write_cstr(&o,"CSysop\r\n");

/*
y        Local postings are allowed
n        No local postings are allowed, only remote ones
m        The group is moderated and all postings must be approved
j        Articles in this group are not kept, but only passed on
x        Articles cannot be posted to this newsgroup
=foo.bar Articles are locally filed into the ``foo.bar'' group
*/

		/* Flags-Zeile: */
		uoio_write_cstr(&o,"F");
		/* L: lesen. + bestellt. P Pflicht (n/a). - verboten (n/a). B einfach bestellen */
		{
			DYN_ENTRY item,*found;
			item.crc=crc32ansi(itg->name,strlen(itg->name));
			found=hhsearch(ghash_handle,&item,DYN_FIND);
			if (found) 
				uoio_write_cstr(&o,"L+");
			else
				uoio_write_cstr(&o,"LB");
			if (*itg->flags=='y'||*itg->flags=='m') { /* very simple */
				if (found) 
					uoio_write_cstr(&o,"S+");
				else
					uoio_write_cstr(&o,"SB");
			}
			uoio_write_cstr(&o,"S-");
		}
		uoio_write_cstr(&o,"P-$-V?\r\n");
		if (*itg->flags=='m')
			uoio_write_cstr(&o,"MO\r\n");
		else
			uoio_write_cstr(&o,"M\r\n");


		/* good enough for the moment */
		if (0==strncmp(itg->name,"maus.",5))
			uoio_write_cstr(&o,"H1:MausNet\r\n");
		else if (0==strncmp(itg->name,"z-netz.",7)
				|| 0==strncmp(itg->name,"t-netz.",7)
				|| 0==strncmp(itg->name,"zer.",4))
			uoio_write_cstr(&o,"H4:Z-Netz\r\n");
		else if (0==strncmp(itg->name,"fido.",5))
			uoio_write_cstr(&o,"H3:FidoNet\r\n");
		else if (0==strncmp(itg->name,"tirka.",6))
			uoio_write_cstr(&o,"H5:MailingListe\r\n");
		else
			/* why bother? Usenet is everything */
			uoio_write_cstr(&o,"H2:Usenet\r\n");
		/* good enough for the moment */
		if (0==strncmp(itg->name,"maus.",5)) {
			if (0==strcmp("maus.talk.english",itg->name))
				uoio_write_cstr(&o,"P2:Englisch\r\n");
			else if (0==strcmp("maus.talk.francais",itg->name))
				uoio_write_cstr(&o,"P3:Franzsisch\r\n");
			else
				uoio_write_cstr(&o,"P1:Deutsch\r\n");
		} else if (0==strncmp(itg->name,"de.",3)
				|| 0==strncmp(itg->name,"z-netz.",7)
				|| 0==strncmp(itg->name,"at.",3)) {
			uoio_write_cstr(&o,"P1:Deutsch\r\n");
		} else if (0==strncmp(itg->name,"comp.",5)
				|| 0==strncmp(itg->name,"news.",5)
				|| 0==strncmp(itg->name,"misc.",5)
				|| 0==strncmp(itg->name,"talk.",5)
				|| 0==strncmp(itg->name,"free.",5)
				|| 0==strncmp(itg->name,"alt.",4)
				|| 0==strncmp(itg->name,"soc.",4)
				|| 0==strncmp(itg->name,"rec.",4)
				|| 0==strncmp(itg->name,"sci.",4)
				|| 0==strncmp(itg->name,"biz.",4)
				|| 0==strncmp(itg->name,"gnu.",4)
				|| 0==strncmp(itg->name,"humanities.",11)
				|| 0==strncmp(itg->name,"uk.",3)) {
			uoio_write_cstr(&o,"P2:English\r\n");
		} else if (0==strncmp(itg->name,"fr.",3)) {
			uoio_write_cstr(&o,"P3:Franzsisch\r\n");
		} else if (0==strncmp(itg->name,"fido.",5)) {
			if (strstr(itg->name,".ger"))
				uoio_write_cstr(&o,"P1:Deutsch\r\n");
			else
				uoio_write_cstr(&o,"P0:Unbekannt\r\n");
		} else
			uoio_write_cstr(&o,"P0:Unbekannt\r\n");

		itg=itg->next;
	}
	uoio_flush(&o);
	crc32ansi_finish(&itg_crc);

	uostr_xdup_cstr(&crc_fname,fname);
	uostr_xadd_mem(&crc_fname,".crc",5);
	fd=open(crc_fname.data,O_WRONLY|O_CREAT|O_TRUNC,0644);
	if (fd==-1) {
		log(LOG_ERR,"can't open %s: %s",crc_fname.data,strerror(errno));
		exit(1);
	}
	uoio_assign_w(&o,fd,write,0);
	str_ulong(numbuf,itg_crc);
	uoio_write_cstr(&o,numbuf);
	uoio_write_cstr(&o,"\n");
	if (-1==uoio_flush(&o)) {
		log(LOG_ERR,"can't write to %s: %s",crc_fname.data,strerror(errno));
		exit(1);
	}
	uostr_freedata(&crc_fname);
}

