/* MausConnect: Nachrichtenformatskonverter fuer das MausNet
   Copyright (C) 1992 1993 1994 1995 Michael Kraehe, Uwe Ohse

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   srfcmail.c: send metamsg as rfc822-mail 
*/

#include "config.h"

#include <sys/types.h>
#ifdef HAVE_SYS_STAT_H
#  include <sys/stat.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <syslog.h>
#include <unistd.h>
#include <errno.h>
#include <ctype.h>
#include <conflib.h>
#include <time.h>
#include <sys/wait.h>

#include "mausconlib.h"
#include "mauscon.h"
#include "metamsg.h"
#include "rfc.h"
#include "srfcmail.h"
#include "msg_io.h"


#ifdef HAVE_LIBRESOLV
#define WANT_DIRECT_SMTP
#endif

#ifdef WANT_DIRECT_SMTP
#include <netinet/in.h>
#include "dns.h"
#endif

/*
   Zwei Methoden werden zur Zeit untersttzt, die erste ist "echtes"
   smtp, die zweite smtp via Batchfile. Das erste ist prinzipiell
   schneller und resourcensparender, das zweite problemloser, da
   smtliche Fehler vom rsmtp gehandhabt werden. 
   Mglicherweise gibt's Dupes, aber das sollte nur dann passieren,
   wenn der smtp-Daemon in ungluecklichen Momenten die 
   Verbindung unterbricht.
 */
#define CRLF "\r\n"


#ifdef WANT_SMTP
static mio_t *local_smtp;
static BOOLEAN local_tried;

static mio_t *direct_smtp;
#ifdef WANT_DIRECT_SMTP
static char direct_site[128];
#endif
#endif

static mio_t *batched_smtp;
static long batchcount;

/* 
 * Aufrufen, wenn es gebraucht wird, d.h., wenn der direkte smtp-Versuch
 * gescheitert ist. Falls auch das nicht klappt, weil das Batchfile
 * nicht geffnet werden kann -> Pech, Ende, Aus. Fr den Gatemodus
 * sollte man das noch mal berdenken, aber fr den Usermodus ist es
 * das beste.
 */
static BOOLEAN 
open_batchfile(void)
{
	uostr_t fname=UOSTR_INIT;
	uostr_t errbuf=UOSTR_INIT;
	size_t pos;
	long i;
	struct stat st;
	uostr_xadd_cstrmulti(&fname,opts.spoolbase,"/",action,"/incoming/mailbatch.",NULL);
	pos=fname.len;
	i=batchcount;
	do
	{
		char num[STR_ULONG];
		uostr_xneed(&fname,STR_ULONG);
		uostr_cut(&fname,pos);
		str_ulong(num,++i);
		uostr_xadd_cstr(&fname,num);
		uostr_x0(&fname);
	} while (stat(fname.data,&st)==0 && st.st_size!=0);
	batchcount=i;
	if (MIO_OK==mio_batch_smtp.open(fname.data,"w",&batched_smtp,&errbuf))
	{
		msglog("  <generic>: <bsmtp> opened batchfile %s", fname.data);
	}
	else
	{
		msglog("  <generic>: <bsmtp> cannot open batchfile %s: %s", 
			fname.data, errbuf.data);
		log(LOG_ERR,_("cannot open batchfile %s: %s"), 
			fname.data, errbuf.data);
		/* Das ist unsere Fallbackmethode - und wenn das schief geht, 
		 * gibt es zur Zeit keine Mglichkeit mehr, Mails zu versenden.
		 * Deshalb geben wir jetzt auf.
		 */
		exit(1);
	}
	return TRUE; /* OK */
}

static void
send_maildir(metamsg_t *msg)
{
	char host[128];
	char num[STR_ULONG];
	uostr_t fn=UOSTR_INIT;
	uostr_t fn2=UOSTR_INIT;
	FILE *f;
	char *p;
	char *pos_tmp;
	struct stat st;
	int r;
	int loop=0;
	int fd=-1;
	char *quark_a_zeile=xentumlauten1(msg->quark_a_zeile);
	str_lwr(quark_a_zeile);
	for (p=quark_a_zeile;*p;p++) {
		if (isspace(*p) || *p=='/')
			*p='_';
	}
	gethostname(host,sizeof(host));
	while (1) { /* for continue und break */
		int count=5;
		loop++;
		if (loop==2) {
			/* create Maildir */
			uostr_xdup_cstr(&fn,opts.quark_user_dir);
			uostr_xadd_cstr(&fn,quark_a_zeile);
			uostr_x0(&fn);
			mkdir(fn.data,0700);

			uostr_cut(&fn,-1);
			uostr_xadd_cstr(&fn,"/Maildir");
			uostr_x0(&fn);
			mkdir(fn.data,0700);

			uostr_cut(&fn,-1);
			uostr_xadd_cstr(&fn,"/new");
			uostr_x0(&fn);
			mkdir(fn.data,0700);

			uostr_cut(&fn,-5);
			uostr_xadd_cstr(&fn,"/cur");
			uostr_x0(&fn);
			mkdir(fn.data,0700);

			uostr_cut(&fn,-5);
			uostr_xadd_cstr(&fn,"/tmp");
			uostr_x0(&fn);
			mkdir(fn.data,0700);
		}
		
		while(1) { /* find unused filename */
			uostr_xdup_cstr(&fn,opts.quark_user_dir);
			uostr_xadd_cstr(&fn,quark_a_zeile);
			pos_tmp=fn.data+fn.len+9;
			uostr_xadd_cstr(&fn,"/Maildir/tmp/");
			str_ulong(num,time(NULL));
			uostr_xadd_cstr(&fn,num);
			uostr_xadd_char(&fn,'.');
			str_ulong(num,getpid());
			uostr_xadd_cstr(&fn,num);
			uostr_xadd_char(&fn,'.');
			uostr_xadd_cstr(&fn,host);
			uostr_x0(&fn);
			r=stat(fn.data,&st);
			if (r==0) {
				if (st.st_mtime>86400) {
					unlink(fn.data);
				}
			} else {
				if (errno==ENOENT)
					break;
			}
			count--;
			if (!count)
				goto fail;
			sleep(2);
		}
		uostr_xdup_uostr(&fn2,&fn);
		memcpy(fn2.data+(pos_tmp-fn.data),"new",3);
		fd=open(fn.data,O_WRONLY | O_EXCL | O_CREAT,0600);
		if (fd==-1) {
			if (loop==1)
				continue; /* try again after creation of maildir */
			goto fail;
		}
		f=fdopen(fd,"w");
		if (!f) {
			goto fail;
		} else {
			fprintf(f, "X-MausConnect-Env-To: %s\n"
					   "X-MausConnect-Env-From: %s\n",
					   msg->env_to, msg->env_from);
			rfc_print_message(msg,f,"\n",0);
			fflush(f);
			if (fsync(fd) == -1)  {
				fclose(f);
				goto fail;
			}
			if (fclose(f))
				goto fail;
		}
		if (-1==link(fn.data,fn2.data)) {
			unlink(fn.data);
			if (loop==1)
				continue; /* try again after creation of maildir */
			goto fail;
		}
		unlink(fn.data);
		uostr_freedata(&fn);
		uostr_freedata(&fn2);
		_exit(0);
	}
  fail:
	uostr_freedata(&fn);
	uostr_freedata(&fn2);
	_exit(1);

}

BOOLEAN
srfcmail (BOOLEAN only_convert, BOOLEAN quark_user, metamsg_t * msg)
{
	uostr_t errbuf=UOSTR_INIT;
	static int atexit_done = 0;
	if (!atexit_done) {
		atexit_done = 1;
		atexit (srfcmail_end_all);
	}
	if (!msg) {
		log (LOG_WARNING, _ ("srfcmail: no message"));
		/* Wenn man nichts versendet, kann man auch keine Fehler machen */
		return TRUE;
	}
	if (!msg->env_to) {
		log (LOG_WARNING, _ ("srfcmail: kein Empfaenger"));
		return FALSE;
	}
	if (only_convert) {
		rfc_print_message (msg, stdout, "\n", 0);
		return TRUE;
	}
	if (quark_user && msg->quark_a_zeile) {
		pid_t pi;
		fflush(NULL); /* need to get a sane state everywhere */
		pi=fork();
		if (pi==0) {
			/* child */
			send_maildir(msg);
			/* not reached */
		} else if (pi!=-1) {
			int st;
			if (pi==waitpid(pi,&st,0) && WIFEXITED(st) && WEXITSTATUS(st)==0) {
				msglog ("  %s: saved to maildir", msg->message_id);
				return TRUE;
			}
		}
		log(LOG_ERR,"can't save message to maildir");
		msglog ("  %s: can't save to maildir",msg->message_id);

		/* fallthrough to other delivery methods. This will create a 
		 * message loop ... but it _should_ not happen anyway, the BBS
		 * should care about access rights.
		 * And anyway, we _have_ to let that message survive anyhow, and
		 * this is our only way
		 */
	}
#ifdef WANT_SMTP
#ifdef WANT_DIRECT_SMTP
	/* Zuerst der Versuch, die Nachricht direkt zu verschicken, d.h.,
	 * ber einen anderen Rechner / anderen SMTP-Server */
	if (opts.direct_smtp && !dont_use_smtp) {
		char *p;
		dns_t *mx;
		dns_t *mxanker;
		int count;
		p = strchr (msg->env_to, '@');
		/* 
		 *  ### MX-Lookup!
		 */
		if (!p)
			goto not_direct;
		count = dns_mx (p + 1, &mx);
		switch (count) {
		case 0:
			/* connect to host: no MX -> use A record */
			mx = XMALLOC (sizeof (dns_t));
			mx->pref = 0;
			mx->name = XSTRDUP (p + 1);
			mx->next = NULL;
			break;
		case DNS_SOFT:
		case DNS_HARD:
			/* DNS_HARD should be bounced, but that's the job
			 * of a real mailer */
			goto not_direct;
		}
		mxanker = mx;
		/* we do have a MX list */
		while (mx) {
			if (direct_smtp && strcasecmp (mx->name, direct_site) != 0) {
				direct_smtp->fn->close (direct_smtp);
				direct_smtp = NULL;
			}
			if (!direct_smtp) {
				if (MIO_OK != mio_smtp.open (mx->name, "w", &direct_smtp, &errbuf)) {
					msglog ("  %s: cannot open %s for %s: %s",
							msg->message_id, mx->name, p + 1, errbuf.data);
					log (LOG_ERR, _ ("%s: cannot open %s for %s: %s"), msg->message_id,
						 mx->name, p + 1, errbuf.data);
					if (mx->next) {
						msglog ("  %s: last MX failed", msg->message_id);
						log (LOG_ERR, _ ("%s: last MX failed"), msg->message_id);
					}
				} else {
					strcpy (direct_site, mx->name);
					msglog ("  %s: opened host %s for %s", msg->message_id, mx->name, p + 1);
				}
			}
			if (direct_smtp) {
				mio_retcode_t ret;
				ret = direct_smtp->fn->export (direct_smtp, msg);
				if (ret == MIO_OK) {
					dns_t *n;
					while (mxanker) {
						free (mxanker->name);
						n = mxanker->next;
						free (mxanker);
						mxanker = n;
					}
					return TRUE;	/* verschickt */
				} else if (ret == MIO_ERROR) {
					/* das sollten wir bouncen */
					dns_t *n;
					msglog ("  %s: cannot send to %s for %s: %s",
							msg->message_id, mx->name, p + 1, direct_smtp->errtext.data);
					log (LOG_ERR, _ ("%s: cannot send to %s for %s: %s"),
						 msg->message_id, mx->name, p + 1, direct_smtp->errtext.data);
					while (mxanker) {
						free (mxanker->name);
						n = mxanker->next;
						free (mxanker);
						mxanker = n;
					}
					direct_smtp->fn->close (direct_smtp);
					direct_smtp = NULL;
					goto not_direct;
				}
			}
			mx=mx->next;
		}
		/* keiner der MX-Hosts hat die Nachricht angenommen */
		{
			dns_t *n;
			while (mxanker) {
				free (mxanker->name);
				n = mxanker->next;
				free (mxanker);
				mxanker = n;
			}
		}
	}
  not_direct:
#endif

	/* Nun der Versuch, die Nachricht an den Hauptserver zu schicken 
	 * (an die eigene Kiste).
	 */
	if (local_smtp == NULL && !local_tried && !dont_use_smtp) {
		local_tried = TRUE;
		if (MIO_OK == mio_smtp.open (opts.mailhost, "w", &local_smtp, &errbuf))
			msglog ("  %s: opened host %s", msg->message_id, opts.mailhost);
		else {
			msglog ("  %s: cannot open %s: %s",
					msg->message_id, opts.mailhost, errbuf.data);
			log (LOG_ERR, _ ("%s: cannot open %s: %s"),
				 msg->message_id, opts.mailhost, errbuf.data);
			local_smtp = NULL;
		}
	}
	if (local_smtp) {
		mio_retcode_t ret;
		ret = local_smtp->fn->export (local_smtp, msg);
		if (ret == MIO_OK)
			return TRUE;		/* verschickt */
	}
#endif
	/* Letzter Versuch, das Batchfile */
	if (!batched_smtp) {
		if (!open_batchfile ())
			return FALSE;
	}
	if (MIO_OK == batched_smtp->fn->export (batched_smtp, msg)) {
		msglog ("  %s: mailbatch %ld", msg->message_id, batchcount);
		return TRUE;
	}
	msglog ("# %s: ERROR %s", msg->message_id, 
		batched_smtp->errtext.data);
	return FALSE;
}

void
srfcmail_end_all(void)
{
#ifdef WANT_SMTP
	if (direct_smtp)
	{
		direct_smtp->fn->close(direct_smtp);
		direct_smtp=NULL;
	}
	if (local_smtp)
	{
		local_smtp->fn->close(local_smtp);
		local_smtp=NULL;
	}
#endif
	if (batched_smtp)
	{
		batched_smtp->fn->close(batched_smtp);
		batched_smtp=NULL;
	}
}
