/*
 * Copyright (C) 1999 Uwe Ohse
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * As a special exception this source may be used as part of the
 * SRS project by CORE/Computer Service Langenbach
 * regardless of the copyright they choose.
 * 
 * Contact: uwe@ohse.de
 */
#include "uotime.h"

/* somewhat faster than mktime, and not influenced by the environment */
uo_sec70_t 
uo_dt2sec70 (uo_datetime_t *tm)
{
	int days;
	int year;
	uo_sec70_t ret;
	int schalt;
	static long days_to_month[] =
	{0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365};

	ret = 3600 * tm->hour + 60 * tm->min + tm->sec;

	year = tm->year;
	if (year % 4)
		schalt = 0;
	else if (year % 400 == 0)
		schalt = 1;
	else if (year % 100 != 0)
		schalt = 1;
	else
		schalt = 0;
	days = (year - 1970) * 365;
	days += ((year - 1969) / 4);
	days -= ((year - 2000) / 100);
	days += ((year - 2000) / 400);
	days += days_to_month[tm->mon];
	days += tm->day-1;
	if (schalt && tm->mon > 1)
		days++;
	ret += 86400 * days;
	return ret;
}

#ifdef TEST
#include <time.h>
int 
main (void)
{
	int i, j;
	struct tm tm;
	uo_sec70_t t1;
	uo_datetime_t dt;
	time_t t2;
	time(&t2);
	tm=*localtime(&t2);


	for (i = 60; i < 400; i++) {
		for (j = 0; j < 12; j++) {
			dt.year = i+1900;
			dt.mon = j;
			dt.day = 1;
			dt.hour = 0;
			dt.min = 0;
			dt.sec = 0;
			tm.tm_year = i;
			tm.tm_mon = j;
			tm.tm_mday = 1;
			tm.tm_hour = 0;
			tm.tm_min = 0;
			tm.tm_sec = 0;
			t1 = uo_dt2sec70 (&dt);
			t2 = mktime (&tm);
			if (t1 != t2)
				printf ("%2d:%2d t1: %lu, t2: %lu, diff %ld\n", i,j,t1, t2, t1 - t2);

		}
	}
	return 0;
}
#endif
